# Writing Tests for Command Line Utilities

## Package Layout

The layout of our hypothetical package is as follows:

```text
mypackage/

    README.md
    setup.py
    requirements.txt

    ...

    mypackagecli/
        __init__.py
        command.py
        tests/
            __init__.py
            test_mypackage.py
            test_mypackagecli.py
            test_utilities.py

```

We structure the module this way to allow both our tests
and our command line utility to be submodules of our 
top-level module/package.

The command line utility's main interface is defined in
`mypackagecli/command.py`. It will define a `main()` entry point function
that takes an arguments list (from stdin by default),
and will use `argparse` or a similar package to parse
the input arguments:

**`mypackagecli/command.py`:**

```python
import sys

...

def main(sysargs = sys.argv[1:]):
    
    ...
```

Defining a command line utility this way allows the 
utility to be called from the command line via 
`subprocess.call()`, or directly as a Python function.
We can write tests using either method.

Likewise, we also include the following in `__init__.py`
of the top-level package to make it easy to change the version number
and command line utility name:

**`mypackagecli/__init__.py`:**

```python
_program = "my-custom-named-cli"
__version__ = "1.0.1beta"
```

Once the package is installed using `setup.py` (see below),
the command line utility will be available from the 
command line (assuming your Python `bin/` is on your path,
see below) as:

```text
$ my-custom-named-cli --help
```


## Test Details in setup.py

The `setup.py` file specifies two keyword arguments:
`test_suite` and `tests_require`. These specify how 
tests should be run, and what packages are required 
to run tests. 

**`setup.py`:**

```python
setup(
        ...
        test_suite = ... ,
        tests_require = ... ,
        ...
)
```

Given the structure above, we also need
to define an entrypoint for the command line
interface in order to be able to use it 
from the command line. We do that with 
the `entry_points` keyword argument:

**`setup.py`:**

```python
setup(
        ...

        entry_points="""
        [console_scripts]
        {program} = cli.command:main
        """.format(program = _program),

        ...
)
```


## Using Nose

We recommend using nose to run unit tests:

**`setup.py`:**

```python
setup(
        ...
        test_suite='nose.collector',
        tests_require=['nose'],
```

Once we specify how to run the test suite, we should be
able to run tests from `setup.py` like so:

```
python setup.py test
```


## Setting `$PATH`

To be able to access the command line utility directly from the 
command line, it msut be on your `$PATH`. The location where 
the command line utility will be installed depends on your 
Python distribution, operating system, and installation line,
but if you install system-wide it will be something like

```text
$HOME/Library/Python/3.6/bin
```

...etc...

## Tests Directory

To write tests, add a `tests/` directory inside the submodule path.

Structure the tests directory as follows:

```text
tests/
    __init__.py
    test_mypackage.py
    ...
    test_utilities.py
```

Here are the details of what each file should look like:

* `__init__.py` is empty

* `test_utilities.py` contains utility methods, 
    particularly methods for capturing stdout 
    and stderr as part of tests (see next section).

* `test_mypackagecli.py` defines objects that extend 
    `unittest.TestCase`. A `TestCase` is simply an object
    that defines test methods (see **Writing Tests** section 
    below).

### Test Utilities 

Here is the contents of the test utilities file.
This defines a method that extracts stdout and 
stderr using `contextlib`.

**`mypackagecli/tests/test_utilities.py`:**

```python
import sys
from contextlib import contextmanager
from io import StringIO

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err
```

## Writing Tests

We cover the contents of `test_mypackagecli.py` to test the 
command line utility. 

The first thing we need in the test file is the import statement
that makes the `main()` method of the command line utility 
availabile to us:

```
from mypackagecli.command import main
```

Note that this command can be run from anywhere,
as it is importing the command line interface's
`main()` function from the `mypackagecli.command` 
submodule.

(In other words, we don't _have_ to structure
our test directory in the project like we did above,
because the tests don't _need_ to live in a 
particular location relative to the package.)

Next we have:

**`mypackagecli/tests/test_mypackagecli.py`:**

```python
# This is the class we will extend to write tests
from unittest import TestCase

class TestPackageCli(TestCase):
    """
    Your docstring here
    """

    def test_something(self):
        """
        Describe your tests!
        """
        self.assertTrue(True)
```

Let's include two more tests. We will discuss these in more detail.

The first test is a test of the command line utility from the actual
command line, using the `subprocess` module. To do this, we need to 
be able to capture stdout and stderr. We do this using the `subprocess.PIPE`
object, which captures stdout and stderr as binary streams, and convert
those to UTF8 using the `decode()` method.

Once we've captured the output, we can use the many assert methods
[implemented by the TestCase class](https://docs.python.org/3/library/unittest.html#assert-methods).


**`mypackagecli/tests/test_mypackagecli.py`:**

```python
# This is how we call the command line utility from the command line
from subprocess import call, Popen, PIPE

...

class TestPackageCli(TestCase):
    
    ...

    def test_using_subprocess(self):
        """
        This illustrates how to run a command using subprocess
        and capture stdout and stderr to examine it as part of 
        the test. The "cwd" keyword (current working directory)
        is optional.
        """
        command = ['/bin/ls','-1']
        p = Popen(command, cwd= ... , stdout=PIPE, stderr=PIPE).communicate()

        p_err = p[0].decode('utf-8').strip()
        p_out = p[1].decode('utf-8').strip()

        # use self.assertIn or self.assertEquals to 
        # test the output of the command 
```

The second way we can test the command line utility is by calling 
the `main()` function from Python directly. Here is an example of
how we do that.

**`mypackagecli/tests/test_mypackagecli.py`:**

```python
# This is how we import the command line utility's Python API
from mypackagecli.command import main

...

class TestPackageCli(TestCase):
    
    ...

    def test_exit_code(self):
        """
        Assert that calling taco ls without arguments or rules
        will result in an exit code -1
        """
        with captured_output() as (out, err):
            main(["do_some_action","--some-argument"])
        output = out.getvalue().strip()
        self.assertIn('foobar',output)
```

The above call to `main()` is equivalent to running the following
command from the command line:

```text
$ my-custom-named-cli do_some_action --some-argument
```

Now suppose we wish to detect whether our command line utility
properly exits with an error code of (for example) -1 when 
invalid arguments are passed to it. We can do that as follows:

```
        with self.assertRaises(SystemExit) as cm:
            with captured_output() as (out, err):
                main(["do_some_action","--some-argument"])
            output = out.getvalue().strip()

        # Exception objects have .error_code attribute
        # SystemExit objects have .code
        # https://stackoverflow.com/a/15672165
        the_exception = cm.exception
        self.assertEqual(the_exception.code, -1)
```



